open Sast
module L = Llvm
module A = Ast
module StringMap = Map.Make(String)


(*translate : Sast.program -> LLvm.module*)
let translate program  =

  (*Decompose program*)
  let (fdecls, stmts) = program in (*List to store function decls, statements, action decls and listener decls*)
   let context = L.global_context() in
    let the_module = L.create_module context "Scolang" in

    (*======= TYPES ========*)
    (*Get types from the context and assigning them to local variables*)
    let function_ptr_type = L.function_type (L.i32_type context) (Array.of_list([]))
    and function_ptr2_type = L.function_type (L.void_type context) (Array.of_list([]))
    and i32_t   = L.i32_type    context
    and i8_t    = L.i8_type     context
    and i1_t    = L.i1_type     context
    and float_t = L.double_type context
    and void_t  = L.void_type   context in
    (*pointer types and contract*)
    let str_t   = L.pointer_type (L.i8_type context)
    and listener_t = L.pointer_type function_ptr_type
    (* and contract_t = L.i32_type context TODO: Get contracts working *)
    and action_t = L.pointer_type function_ptr2_type
    in

    (*Return the LLVM type for Scolang type*)
    let ltype_of_primitive = function
      A.Int   -> i32_t
    | A.Bool  -> i1_t
    | A.String -> str_t
    | A.Float -> float_t
    | A.Void  -> void_t
    | A.Contract -> i32_t
    | A.Listener -> listener_t
    | A.Action -> action_t
    in

    (* TODO!!!! FIGURE OUT TYPE HEIRARCHY AND FIX IT*)
    (*=====================*)

    (* Function to convert from typ type to primitive type *)
    let prim_of_dt = function
      | A.Primitive(p)  -> p
      (* | A.Array(p, _)   -> p TODO: Arrays *)
    in


    (*======= EXTERNAL FUNCTIONS ========*)

    let print_t : L.lltype =
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
    let print_func : L.llvalue =
      L.declare_function "print" print_t the_module in

    let printf_t : L.lltype =
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
    let printf_func : L.llvalue =
      L.declare_function "printf" printf_t the_module in

    let bindval_t : L.lltype =
      L.function_type i32_t [|L.pointer_type function_ptr_type; L.pointer_type function_ptr2_type|] in
    let bindval_func : L.llvalue =
      L.declare_function "bind_val" bindval_t the_module in

    let query_t : L.lltype =
      L.var_arg_function_type i32_t [| L.pointer_type i8_t  |] in
    let query_func : L.llvalue =
      L.declare_function "query" query_t the_module in

    let webhook_t : L.lltype =
      L.var_arg_function_type i1_t [| i32_t |] in
    let webhook_func : L.llvalue =
      L.declare_function "webhook" webhook_t the_module in

    let system_call_t : L.lltype =
      L.var_arg_function_type i32_t [| L.pointer_type i8_t |] in
    let system_call_func : L.llvalue =
      L.declare_function "system_call" system_call_t the_module in

    (*===================================*)

    (*======= GLOBAL VARIABLES ==========*)
    (* TODO Figure out this matching here*)
    (* TODO Actually store global values in the global var map *)

    (*Function to check whether something is an initialisation for global var*)
    let is_global_initialization stmt = match stmt with
      SVarInit(_,_,_) -> true
      |_ -> false
    in

    let get_global_name stmt = match stmt with
      SVarInit(t,n,_) ->  prim_of_dt t,n
      | _ -> failwith (" Globab extraction failing  " ^ string_of_sstmt stmt)
    in

    let globals_list = List.filter is_global_initialization stmts
    in
    let globals = List.map get_global_name globals_list
    in
    (*===================================*)

    (*======= EXPOSED STATEMENTS ==========*)
    let is_exposed_statement stmt = match stmt with
      SVarInit(_,_,_) -> true
      | SLisActInit(_,_,_) -> false
      | _ -> true
    in

    let pure_statements = List.filter is_exposed_statement stmts
    in
    (*===================================*)

    (* ======= Create an array of Listeners and Action Functions ======= *)
    let all_actions =
      let pop_actions action_list stmt = match stmt with
        SLisActInit(t, n, s) ->
            if A.Action = prim_of_dt t then
            {
              styp = A.Primitive(A.Void);
              sfname = n;
              sformals = [];
              sbody = s
            } :: action_list
            else (action_list)
        | _ -> action_list
      in  List.fold_left pop_actions [] stmts
    in

    let all_listeners =
      let pop_listeners listener_list stmt = match stmt with
        SLisActInit(t, n, s) ->
            if A.Listener = prim_of_dt t then
            {
              styp = A.Primitive(A.Int);
              sfname = n;
              sformals = [];
              sbody = s
            } :: listener_list
            else (listener_list)
        | _ -> listener_list
      in  List.fold_left pop_listeners [] stmts
    in

    (* TODO: Scan for actions and listeners inside of functions *)

    (* Create a string map of actions *)
    let action_decls: (L.llvalue * sfunc_decl) StringMap.t =
      let function_decl map action =
        let name = action.sfname in
        let ftype = L.function_type (ltype_of_primitive A.Void) (Array.of_list([])) in
        StringMap.add name (L.define_function name ftype the_module, action) map
      in
      List.fold_left function_decl StringMap.empty all_actions
    in

    (* Create a string map of listeners *)
    let listener_decls: (L.llvalue * sfunc_decl) StringMap.t =
      let function_decl map listener =
        let name = listener.sfname in
        let ftype = L.function_type (ltype_of_primitive A.Int) (Array.of_list([])) in
        StringMap.add name (L.define_function name ftype the_module, listener) map
      in
      List.fold_left function_decl StringMap.empty all_listeners
    in

    (*======= CREATING MAIN FUNCTION =================*)
    (*We're now going to create a list of functions from our fdecls and build our faux main with pure statements*)
    let functions =
      let main_fdecl = {
          styp = A.Primitive(A.Int);
          sfname = "main";
          sformals = [];
          sbody = List.rev(SReturn(A.Int, SIntLit(0)) :: List.rev(pure_statements))
        }
      in main_fdecl :: fdecls
    in
    (*===================================*)

    (*=========== GLOBAL VARS ================*)
    (* Store the global variables in a string map *)
    (*TODO LOOK AT LLVALUE STRINGMAP*)
    let global_vars : L.llvalue StringMap.t=
      let global_var map (t, n) =
        let init = match t with
          A.Float -> L.const_float (ltype_of_primitive t) 0.0
          | A.String -> L.const_null (ltype_of_primitive t)
          | A.Void -> L.const_null (ltype_of_primitive t)
          | A.Action -> L.const_null (ltype_of_primitive t)
          | A.Listener -> L.const_null (ltype_of_primitive t)
          | _ -> L.const_int (ltype_of_primitive t) 0
        in StringMap.add n (L.define_global n init the_module) map
      in List.fold_left global_var StringMap.empty globals
    in
    (*========================================*)


    (*=========== FUNCTIONS AND FUNC. POPULATION ================*)
    let function_decls : (L.llvalue * sfunc_decl) StringMap.t =
      let function_decl map fdecl =
        let name = fdecl.sfname
        and formal_types = Array.of_list (List.map (fun (t, _) -> ltype_of_primitive (prim_of_dt t)) fdecl.sformals)
          in
        let ftype = L.function_type (ltype_of_primitive (prim_of_dt fdecl.styp)) formal_types in
        StringMap.add name (L.define_function  name ftype the_module, fdecl) map
      in
      List.fold_left function_decl StringMap.empty functions
    in

    let build_function_body decl_list fdecl =
      let (the_function, _) = StringMap.find fdecl.sfname decl_list in
      let builder = L.builder_at_end context (L.entry_block the_function) in
      let int_format_str = L.build_global_stringptr "%d\n" "fmt" builder
      and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder
    in

    let local_vars =
      let add_formal var_map (formal_type, formal_name) param = L.set_value_name formal_name param;
        let local = L.build_alloca (ltype_of_primitive (prim_of_dt formal_type)) formal_name builder in
          ignore(L.build_store param local builder);
          StringMap.add formal_name local var_map
        and add_local map (formal_type, formal_name) =
        let local_var = L.build_alloca (ltype_of_primitive (prim_of_dt formal_type)) formal_name builder in
        StringMap.add formal_name local_var map
      in

      let formals = List.fold_left2 add_formal StringMap.empty fdecl.sformals
        (Array.to_list (L.params the_function)) in

      let function_locals =
        let get_locals_from_fbody function_body =
          let get_vdecl locals_list stmt = match stmt with
            SVarInit(typ, string, _) ->
            if( fdecl.sfname = "main")
              then (typ, string) :: locals_list
              else (typ, string) :: locals_list
            | _ -> locals_list
          in List.fold_left get_vdecl [] function_body
        in get_locals_from_fbody fdecl.sbody
      in List.fold_left add_local formals function_locals
    in

    let lookup n = try StringMap.find n local_vars
      with Not_found -> StringMap.find n global_vars
    in

    let lookup_actions n = try StringMap.find n action_decls
      with Not_found -> raise (Failure ("searched for an undefined action " ^ n))
    in

    let lookup_listeners n = try StringMap.find n listener_decls
      with Not_found -> raise (Failure ("searched for an undefined listener " ^ n))
    in

    (*===================================*)


    let rec expr builder ((_, e) : sexpr)  = match e with
      SIntLit i -> L.const_int i32_t i
    | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
    | SFloatLit l -> L.const_float float_t l
    | SStringLit s -> L.build_global_stringptr s "str" builder
    | SVar s -> L.build_load (lookup s) s builder
    | SBinop (e1, op, e2) ->
      let e1' = expr builder e1 and
      e2' = expr builder e2 in
      (match op with
        A.Lt  -> L.build_icmp L.Icmp.Slt
        | A.Lte -> L.build_icmp L.Icmp.Sle
        | A.Gt ->  L.build_icmp L.Icmp.Sgt
        | A.Gte -> L.build_icmp L.Icmp.Sge
        | A.Eq ->  L.build_icmp L.Icmp.Eq
        | A.NotEq ->  L.build_icmp L.Icmp.Ne
        | A.And -> L.build_and
        | A.Or -> L.build_or
        | A.Plus ->  L.build_add
        | A.Minus ->  L.build_sub
        | A.Times ->  L.build_mul
        | A.Divide ->   L.build_sdiv
        | A.Modulo ->  L.build_srem
        | A.Band ->  L.build_and
        | A.Bor ->  L.build_or
        | A.Bright ->  L.build_lshr
        | A.Bleft ->  L.build_shl
        | A.Bxor  ->   L.build_xor
      ) e1' e2' "tmp" builder
    | SUnop (op, e) -> let e' = expr builder e in
        ( match op with A.Not -> L.build_not) e' "tmp" builder
    | SNoexpr -> L.const_int i32_t 1
    | SAsn (s,e) -> let e' = expr builder e in
      ignore(L.build_store e' (lookup s) builder); e'
    (* | SArrayIndex(v, i)  TODO: Arrays *)
    | SInnerExpr(e) -> expr builder e
    | SCall("print", [e]) -> L.build_call print_func [| (expr builder e) |] "print" builder
    | SCall("printi", [e]) -> L.build_call printf_func [| int_format_str ; (expr builder e) |] "printf" builder
    | SCall ("printf", [e]) -> L.build_call printf_func [| float_format_str ; (expr builder e) |] "printf" builder
    | SCall ("webhook", [e]) -> L.build_call webhook_func [| (expr builder e) |] "webhook" builder
    | SCall ("system_call", [e]) -> L.build_call system_call_func [| (expr builder e) |] "system_call" builder
    | SCall("query", [e]) -> L.build_call query_func [| (expr builder e)|] "query" builder
    | SCall(f, args) ->
        let (fdef, fdecl) = StringMap.find f function_decls in
        let llargs = List.rev (List.map (expr builder)(List.rev args)) in
        let result = (match fdecl.styp with
             A.Primitive(A.Void) -> "" | _ -> f^"_result") in
            L.build_call fdef (Array.of_list llargs) result builder in

        let add_terminal builder instr =
          match L.block_terminator (L.insertion_block builder) with
    	     Some _ -> ()
          | None -> ignore (instr builder) in

    let rec stmt builder = function
      | SVarInit (_, v, e) -> let e' = expr builder e in
                              ignore (L.build_store e' (lookup v) builder); builder
      | SBlock sl -> List.fold_left stmt builder sl
      | SExpr e -> ignore(expr builder e); builder
      | SBind (v1, v2) ->
        let (action_llvalue, _) = lookup_actions v2
        and (listener_llvalue, _) = lookup_listeners v1 in
        ignore(L.build_call bindval_func [|listener_llvalue; action_llvalue|] "bind_val" builder); builder
      | SReturn e -> ignore(match fdecl.styp with
              A.Primitive(A.Void) -> L.build_ret_void builder |
              _ -> L.build_ret (expr builder e) builder);
              builder
      | SIf (predicate, then_stmt, else_stmt) ->
          let bool_val = expr builder predicate in
        	let merge_bb = L.append_block context "merge" the_function in
          let build_br_merge = L.build_br merge_bb in (* partial function *)

        	let then_bb = L.append_block context "then" the_function in
        	 add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
        	   build_br_merge;

        	let else_bb = L.append_block context "else" the_function in
        	 add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
        	   build_br_merge;

        	 ignore(L.build_cond_br bool_val then_bb else_bb builder);
        	 L.builder_at_end context merge_bb

      | SWhile (predicate, body) ->
        let pred_bb = L.append_block context "while" the_function in
        ignore(L.build_br pred_bb builder);

        let body_bb = L.append_block context "while_body" the_function in
        add_terminal (stmt (L.builder_at_end context body_bb) body)
        (L.build_br pred_bb);

        let pred_builder = L.builder_at_end context pred_bb in
        let bool_val = expr pred_builder predicate in
        let merge_bb = L.append_block context "merge" the_function in
        ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
        L.builder_at_end context merge_bb

      | SFor (e1, e2, e3, body) -> stmt builder ( SBlock [SExpr e1; SWhile(e2, SBlock [body; SExpr e3])])
      | SResolve e -> ignore(L.build_ret (expr builder e) builder); builder
      | SReject -> ignore(L.build_ret_void builder); builder
      | SBreak -> ignore(L.build_ret_void builder); builder
      | SContinue -> builder
      | SLisActInit(_, _, _) -> builder (* This case is already handled *)
    in
    let builder = stmt builder (SBlock fdecl.sbody) in

    (* Add a return if the last block falls off the end *)
    add_terminal builder (match fdecl.styp with
        A.Primitive(A.Void) -> L.build_ret_void
      | A.Primitive(A.Float) -> L.build_ret (L.const_float float_t 0.0)
      | t -> L.build_ret (L.const_int (ltype_of_primitive (prim_of_dt t)) 0))
    in
    List.iter (build_function_body function_decls) functions;
    List.iter (build_function_body action_decls) all_actions;
    List.iter (build_function_body listener_decls) all_listeners;
    the_module
