%{
open Ast
%}

%token INT FLOAT BOOL STRING VOID CON ACT LIS EQ NOTEQ RESOLVE REJECT
%token PLUS MINUS TIMES DIVIDE MODULO EOF OPENSQ CLOSESQ OPENPAR CLOSEPAR OPENBLOCK CLOSEBLOCK SEMICOLON COMMA
%token ARRDEC ASSIGN NULL AND OR
%token BAND BOR BRIGHT BLEFT BXOR
%token LT LTE GT GTE
%token NOT
%token BIND
%token FOR WHILE IF ELSE BREAK CONTINUE RETURN
%token<string> VARIABLE
%token<bool> BOOL_LIT
%token<int> INT_LIT
%token<string> STRING_LIT
%token<float> FLOAT_LIT

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left AND OR
%left BAND BOR BXOR
%left EQ NOTEQ
%left LT LTE GT GTE
%left BLEFT BRIGHT
%left PLUS MINUS
%left TIMES DIVIDE MODULO
%right NOT
%right BIND ARRDEC


%start start_state
%type<Ast.program> start_state
%%

start_state:
  decls EOF { $1 }

decls:
   /* nothing */ { [], [] }
 | decls statement { fst $1, (snd $1 @ [$2]) }
 | decls fdecl { (fst $1 @ [$2]), snd $1 }

fdecl:
  vartype VARIABLE OPENPAR formals_opt CLOSEPAR OPENBLOCK stmt_list CLOSEBLOCK
    { { typ = $1;
      fname = $2;
      formals = $4;
      body = List.rev $7 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    vartype VARIABLE                   { [($1,$2)] }
  | formal_list COMMA vartype VARIABLE { ($3,$4) :: $1 }

stmt_list:
   /* nothing */  { [] }
  | stmt_list statement { $2 :: $1 }

statement:
  | expr SEMICOLON { Expr($1) }

  /* control flow */
	| FOR OPENPAR expr SEMICOLON expr SEMICOLON expr CLOSEPAR statement { For($3, $5, $7, $9) }
  | OPENBLOCK stmt_list CLOSEBLOCK { Block(List.rev $2) }
  | WHILE OPENPAR expr CLOSEPAR statement { While($3, $5) }
  | IF OPENPAR expr CLOSEPAR statement %prec NOELSE { If($3, $5, Block([])) }
  | IF OPENPAR expr CLOSEPAR statement ELSE statement    { If($3, $5, $7) }
  | BREAK { Break }
	| CONTINUE { Continue }

  /* Return */
  | RETURN expr SEMICOLON { Return($2) }
  | RETURN SEMICOLON { Return(NoExpr) }

  /* Variable assignment */
  | vartype VARIABLE SEMICOLON { VarInit($1, $2, NoExpr) }
  | vartype VARIABLE ASSIGN expr SEMICOLON { VarInit($1, $2, $4) }
  | vartype VARIABLE ASSIGN OPENBLOCK stmt_list CLOSEBLOCK SEMICOLON { LisActInit($1, $2, (List.rev $5)) }

  /* Binding */
  | VARIABLE BIND VARIABLE SEMICOLON { Bind($1, $3) }

  /* Completion */
  | RESOLVE expr SEMICOLON { Resolve($2) }
  | REJECT SEMICOLON { Reject }

expr:
	| BOOL_LIT	{ BoolLit($1) }
	| INT_LIT	{ IntLit($1) }
  | FLOAT_LIT	{ FloatLit($1) }
  | STRING_LIT	{ StringLit($1) }

  /*Vars*/
	| VARIABLE  { Var($1) }

	/*Logical*/
  | expr LT expr { Binop($1, Lt , $3) }
  | expr LTE expr { Binop($1, Lte , $3) }
  | expr GT expr 	{ Binop($1, Gt , $3) }
  | expr GTE expr { Binop($1, Gte , $3) }
  | expr EQ expr { Binop($1, Eq , $3) }
  | expr NOTEQ expr { Binop($1, NotEq ,$3) }
  | expr AND expr { Binop($1, And, $3) }
  | expr OR expr { Binop($1, Or, $3) }

	/*Unary*/
  | NOT expr { Unop(Not, $2) }

	/*Arithmetic*/
  | expr PLUS expr { Binop($1, Plus, $3) }
  | expr MINUS expr { Binop($1, Minus, $3) }
  | expr TIMES expr { Binop($1, Times, $3) }
  | expr DIVIDE expr { Binop($1, Divide, $3) }
  | expr MODULO expr { Binop($1, Modulo, $3) }

  /*BITWISE*/
  | expr BAND expr { Binop($1, Band, $3) }
  | expr BOR expr { Binop($1, Bor, $3) }
  | expr BRIGHT expr { Binop($1, Bright, $3) }
  | expr BLEFT expr { Binop($1, Bleft, $3) }
  | expr BXOR expr { Binop($1, Bxor, $3) }

  /* Variable Assignment */
  | VARIABLE ASSIGN expr { Asn($1, $3) }
  /* | VARIABLE OPENSQ INT_LIT CLOSESQ { ArrayIndex($1, $3) } TODO: Arrays */

  /*Functional for contract binding & calling */
  /* | VARIABLE ASSIGN VARIABLE BIND VARIABLE { Bind($1, $3) } TODO: Allow Contract variables */
  | VARIABLE OPENPAR actuals_opt CLOSEPAR { Call($1, $3) }

  /*parentheses*/
  | OPENPAR expr CLOSEPAR { InnerExpr($2) }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }

/*variable and array types*/
vartype:
  | datatype { Primitive($1) }
  /* | datatype OPENSQ INT_LIT CLOSESQ { Array($1, $3) } TODO: Arrays */

datatype:
  INT { Int }
  | FLOAT { Float }
  | STRING { String }
  | BOOL { Bool }
  | LIS { Listener }
  | ACT { Action }
  | CON { Contract }
  | VOID { Void }
