open Ast

type sexpr = primitive * sx
and sx =
	  SIntLit of int
  | SBoolLit of bool
  | SFloatLit of float
  | SStringLit of string
	| SVar of string
  | SBinop of sexpr * op * sexpr
  | SUnop of uop * sexpr
  | SAsn of string * sexpr
	| SCall of string * sexpr list
  (* | SArrayIndex of string * int TODO: Arrays *)
  | SInnerExpr of sexpr
	| SNoexpr

type sstatement =
		SExpr of sexpr
  | SFor of sexpr * sexpr * sexpr * sstatement
  | SWhile of sexpr * sstatement
  | SIf of sexpr * sstatement * sstatement
	| SBlock of sstatement list
	| SVarInit of typ * string * sexpr
	| SLisActInit of typ * string * sstatement list
  | SBreak
  | SContinue
	| SBind of string * string
  | SReturn of sexpr
	| SResolve of sexpr
  | SReject


type sfunc_decl =
	{
  	styp:  typ;
    sfname: string;
    sformals: bind list;
    sbody: sstatement list;
	}

type sprogram = sfunc_decl list

(* Pretty-printing functions *)

let rec string_of_sexpr (t, e) =
  "(" ^ string_of_primitive t ^ " : " ^ (match e with
    SIntLit(l) -> string_of_int l
  | SBoolLit(true) -> "true"
  | SBoolLit(false) -> "false"
  | SStringLit(s) -> s
  | SFloatLit(l) -> string_of_float l
  | SVar(s) -> s
  | SBinop(e1, o, e2) ->
      string_of_sexpr e1 ^ " " ^ string_of_op o ^ " " ^ string_of_sexpr e2
  | SUnop(o, e) -> string_of_uop o ^ string_of_sexpr e
  | SAsn(v, e) -> v ^ " = " ^ string_of_sexpr e
  | SCall(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_sexpr el) ^ ")"
  | SNoexpr -> ""
	(* | SArrayIndex(v, i) -> v ^ "[" ^ string_of_int i ^ "]" TODO: Arrays *)
	| SInnerExpr(e) -> "(" ^ string_of_sexpr e ^ ")"
          ) ^ ")"

let rec string_of_sstmt = function
    SBlock(stmts) ->
      "{\n" ^ String.concat "" (List.map string_of_sstmt stmts) ^ "}\n"
  | SExpr(expr) -> string_of_sexpr expr ^ ";\n";
  | SReturn(expr) -> "return " ^ string_of_sexpr expr ^ ";\n";
  | SIf(e, s, SBlock([])) ->
      "if (" ^ string_of_sexpr e ^ ")\n" ^ string_of_sstmt s
  | SIf(e, s1, s2) ->  "if (" ^ string_of_sexpr e ^ ")\n" ^
      string_of_sstmt s1 ^ "else\n" ^ string_of_sstmt s2
  | SFor(e1, e2, e3, s) ->
      "for (" ^ string_of_sexpr e1  ^ " ; " ^ string_of_sexpr e2 ^ " ; " ^
      string_of_sexpr e3  ^ ") " ^ string_of_sstmt s
  | SWhile(e, s) -> "while (" ^ string_of_sexpr e ^ ") " ^ string_of_sstmt s
	| SVarInit(t, v, e) -> string_of_typ t ^ " " ^ v ^ " " ^ string_of_sexpr e
	| SLisActInit(t, v, s) -> string_of_typ t ^ " " ^ v ^ " " ^ (String.concat "" (List.map string_of_sstmt s))
	| SBind(v1, v2) -> v1 ^ " -> " ^ v2
	| SBreak -> "break\n"
	| SContinue -> "continue\n"
	| SReject -> "reject\n"
	| SResolve(e) -> "resolve" ^ string_of_sexpr e ^ "\n"

let string_of_sfdecl fdecl =
  string_of_typ fdecl.styp ^ " " ^
  fdecl.sfname ^ "(" ^ String.concat ", " (List.map snd fdecl.sformals) ^
  ")\n{\n" ^
  String.concat "" (List.map string_of_sstmt fdecl.sbody) ^
  "}\n"

let string_of_sprogram (functions, statements) =
  String.concat "" (List.map string_of_sfdecl functions) ^ "\n" ^
	String.concat "\n" (List.map string_of_sstmt statements)
